# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from cmf.audit.impalad.impalad_inotify_event_processor import ImpaladINotifyEventProcessor
from cmf.audit.impalad.impalad_audit_logs import ImpaladAuditLogs
from cmf.audit.impalad.impalad_audit_sender import ImpaladAuditSender
from cmf.audit.impalad.wal import Wal
from pyinotify import WatchManager, ThreadedNotifier
from threading import Thread, Event
import cmf.audit
import logging
import pyinotify
import time
import itertools
import sys

LOG = logging.getLogger( "impalad_audit_plugin" )

class ImpaladAuditPlugin(object):
  """
  Audit plugin to tail the Impalad audit log and use that to pass audit events
  to the audit server. Takes care of log rotation and resuming from previous
  state in case of agent restart.
  """
  def __init__(self):
    """
    When "refresh" method is called we will create new thread to monitor impalad
    audit log file. Following instance variables are used:
    1. _audit_thread variable will hold the instance of the main thread we will
       create. This thread will go in a while(True) loop to process all incoming
       audit events.
    2. _stop_audit_thread boolean flag will be used to let this thread know that
       no further processing is needed and it should shutdown.
    3. _audit_thread_event is a python Event class that is used to poll for
       a new audit event, and is also used to notify the thread that it should
       exit event processing loop and shutdown.
    4. _audit_logs: Maintains a list of all audit logs in the impalad audit log
       directory. The logs are sorted by their timestamp (each impalad log
       contains timestamp as part of its filename)
    5. _current_audit_log: The current audit log that is being processed.
    """
    self._audit_thread = None
    self._stop_audit_thread = False
    self._audit_thread_event = Event()
    self._audit_logs = None
    self._current_audit_log = None

  def __str__(self, *args, **kwargs):
    return "Audit plugin for Impalad"

  def refresh(self, conf_dir, audit_config_parser = cmf.audit.parse_audit_configuration):
    """
    Parses that audit configuration file to get the location of audit server and
    audit log file. If auditing is enabled it starts a thread that then reads
    the impalad audit log and passes audit events to the audit server.
    """
    try:
      LOG.info("Refreshing %s" % (self,))

      # stop existing auditing thread
      if self._audit_thread is not None:
        self.stop_audit_monitoring()
        self._audit_thread = None

      audit_conf = audit_config_parser(conf_dir, "Impalad")
      if audit_conf is not None:
        self._audit_thread = Thread(target = self._monitor_impala_audit,
                                    name = "Impalad-Audit-Plugin",
                                    kwargs = {'conf' : audit_conf})
        self._audit_thread.setDaemon(True)
        self._audit_thread.start()
    except:
      LOG.exception("Exception caught when trying to refresh %s" % (self,))

  def _monitor_impala_audit(self, conf):
    """
    This method runs in a separate thread and takes care of monitoring impala
    audit log and then generating events from it.
    The method sets up pyinotify system and log reader then invokes
    _process_events method which is the main loop for processing events.
    Once _process_events method returns, this method cleans stops pyinotify and
    closes log file and exits, which terminates this thread.
    """
    wal_file = None
    event_sender = None
    notifier = None
    try:
      # Step 1: Setup audit log file
      wal_file = Wal(conf.log_dir())
      file_name, offset = wal_file.get_file_info()
      LOG.info("Got file %s from wal." % (file_name, ))
      self._audit_logs = ImpaladAuditLogs(conf.log_dir())
      self._current_audit_log = self._audit_logs.get_log(file_name)
      self._current_audit_log.seek(offset)
      LOG.info("Starting with audit log %s" %
                (self._current_audit_log.get_file_name(), ))

      # Step 2: Setup pyinotify
      audit_event_processor = ImpaladINotifyEventProcessor(audit_logs=self._audit_logs)
      wm = WatchManager()
      notifier = ThreadedNotifier(wm, audit_event_processor)
      notifier.setDaemon(True)
      notifier.start()
      wm.add_watch(conf.log_dir(), pyinotify.IN_CREATE | pyinotify.IN_DELETE | pyinotify.IN_MOVED_FROM)

      # Step 3: process reported events in infinite loop, stop only when
      # stop_audit_monitoring method is called.
      event_sender = ImpaladAuditSender(conf.server_url(), conf.service_name())
      LOG.info("Starting auditing event processing loop.")
      self._process_events(conf, event_sender, wal_file)
    except:
      LOG.exception("Unexpected error when starting auditing")

    # Can't put this in finally as python 2.4 can't have except and finally
    # in same statement
    if wal_file is not None:
      wal_file.close()
    if event_sender is not None:
      event_sender.close()
    if notifier is not None:
      notifier.stop()
    if self._current_audit_log is not None:
      self._current_audit_log.close()

    LOG.info("Finished auditing event processing loop.")

  def _process_events(self, conf, event_sender, wal_file):
    """
    Main loop to process audit events. It processes "batch_size" audit events
    or whatever # of events arrive in a configurable time period. This method
    returns when _stop_audit_thread boolean is set. This flag indicates that we
    no longer need to process audit events and should stop.
    """
    timeout = conf.batch_timeout()
    LOG.info("Event poll interval: %d" % (timeout,))
    # Now start the loop to process events forever
    while True:
      self._read_and_send_events(conf, event_sender, wal_file)
      self._audit_thread_event.wait(timeout)
      self._audit_thread_event.clear() # clear flag so we can wait again
      if self._stop_audit_thread:
        break

  def _read_and_send_events(self, conf, event_sender, wal_file):
    """
    This method reads a "batch_size" audit events and then tries to send them
    to the audit server. In case of a failure the method keeps on retrying
    forever till it succeeds.
    If no new audit event is found in the log, the method checks to see if the
    log file has rolled over, in which case it opens the new log file and starts
    reading (and sending) events from that file.
    """
    timeout = conf.batch_timeout()
    batch_size = conf.batch_min_size()

    while True:
      current_file_name = self._current_audit_log.get_file_name()
      next_file_name = self._audit_logs.get_next_log_file_name(current_file_name)

      # process events in batches, read_audit_log method returns a tuple:
      # filename, offset, audit_event
      # The audit_event is a line in audit log that is json parsed and is used
      # to send event to audit server. filename and offset are saved to the
      # WAL file
      audit_queue = None
      if self._current_audit_log.exists():
        audit_queue = list(itertools.islice(
                                     self._current_audit_log.read_audit_log(),
                                     0, batch_size))
      if audit_queue:
        LOG.debug("Fetched %d audits from audit log." % (len(audit_queue),))
        while not event_sender.send_audit_events(audit_queue):
          LOG.debug("Unable to send audits to nav server. Will try again.")
          time.sleep(timeout)
          if self._stop_audit_thread:
            break
        filename, offset, _ = audit_queue.pop()
        wal_file.append(filename, offset)
      elif next_file_name is not None:
        LOG.info("Done processing audit log %s, switching to next log %s." %
                 (current_file_name, next_file_name))
        self._current_audit_log.close()
        self._current_audit_log = self._audit_logs.get_log(next_file_name)
      else:
        break

  def stop_audit_monitoring(self):
    """
    Sets a flag to stop audit monitoring, which results in _audit_thread coming
    out of its main loop.
    """
    self._stop_audit_thread = True
    self._audit_thread_event.set()

def main():
  def mock_parse_audit_configuration(conf_dir, module_name):
    return MockAuditConf()

  class MockAuditConf(object):
    def log_dir(self):
      return "/tmp/impala-audit"

    def server_url(self):
      return "http://localhost:7186"

    def service_name(self):
      return "IMPALA-1"

    def batch_min_size(self):
      return 4

    def batch_timeout(self):
      return 5

  audit_plugin = ImpaladAuditPlugin()
  audit_plugin.refresh("ignored", mock_parse_audit_configuration)

if __name__ == "__main__":
  main()
