-- Copyright (c) 2012 Cloudera, Inc. All rights reserved.
--
-- Version 43 of the CMF schema. Migration code to be applied on top of the
-- previous version.
--

update SCHEMA_VERSION set VERSION=43;

-- Modifications for a generic config container and for host configs container

--
-- Create a new table to store generic config containers
--
create table CONFIG_CONTAINERS (
    CONFIG_CONTAINER_ID int8 not null,
    OPTIMISTIC_LOCK_VERSION int8 not null,
    CONFIG_TYPE varchar(255) not null unique,
    CONFIG_REVISION_ID int8,
    primary key (CONFIG_CONTAINER_ID)
);

--
-- Create a new column in CONFIG_REVISIONS to store config container id
--
alter table CONFIG_REVISIONS
      add column CONFIG_CONTAINER_ID int8;

--
-- Create a new column in HOSTS to store config container id
--
alter table HOSTS
      add column CONFIG_CONTAINER_ID int8;

--
-- Create a new column in CONFIGS to store config container id
--
alter table CONFIGS
      add column CONFIG_CONTAINER_ID int8;

--
-- Add a column in the AUDITS table to refer to a config container to
-- facilitate auditing of updates to configs 'own'-ed by the config container.
--
alter table AUDITS
      add column CONFIG_CONTAINER_ID int8;

--
-- Drop obsolete foreign key constraints
--
alter table CONFIG_REVISIONS
      drop constraint FK_CONFIG_REVISION_HOST;

alter table HOSTS
      drop constraint FK_HOST_CONFIG_REVISION;

--
-- Clean up unwanted foreign key references and columns from the config revision
-- table.
-- Remove existing config revisions that refer to hosts since we no
--   longer have config revisions associated with hosts.
-- Drop the host id column
--
delete from CONFIG_REVISIONS where HOST_ID is not null;
alter table CONFIG_REVISIONS
      drop column HOST_ID;

--
-- Clean up unwanted columns from the hosts table.
-- Drop the config revision id column
--
alter table HOSTS
      drop column CONFIG_REVISION_ID;

--
-- Insert a config container to track host configs.
--
insert into CONFIG_CONTAINERS (CONFIG_CONTAINER_ID, OPTIMISTIC_LOCK_VERSION, CONFIG_TYPE)
       values (1, 1, 'ALL_HOSTS');

-- As preparation to  insert a base revision for existing config containers,
-- create a sequence that can be used for auto incrementing the REVISION_ID
-- field in the CONFIG_REVISIONS table.
-- The sequence is initialized to the maximum value of REVISION_ID from the
-- CONFIG_REVISIONS table.
-- After inserting base revisions for hosts, the sequence is dropped.
--
create sequence CONFIG_REVISIONS_REVISION_ID_SEQ;

select setval('CONFIG_REVISIONS_REVISION_ID_SEQ', max(REVISION_ID)) from CONFIG_REVISIONS;

alter table CONFIG_REVISIONS
      alter column REVISION_ID set default nextval('CONFIG_REVISIONS_REVISION_ID_SEQ');

--
-- Insert a base revision that is visible for each existing config container.
-- After inserting revisions for each config container, the sequence is dropped.
--
insert into CONFIG_REVISIONS (CONFIG_CONTAINER_ID, CREATED_INSTANT, MESSAGE, VISIBLE)
       select CONFIG_CONTAINER_ID, extract(epoch FROM now()) * 1000, 'Base revision', 1
       from CONFIG_CONTAINERS;

alter table CONFIG_REVISIONS
       alter column REVISION_ID drop default;

drop sequence CONFIG_REVISIONS_REVISION_ID_SEQ;

--
-- Link each config container to its current config revision.
--
-- Note that UPDATE doesn't allow aliases in older versions of Postgres, so
-- they aren't used here.
update CONFIG_CONTAINERS
       set CONFIG_REVISION_ID=
       (select REVISION_ID
       from CONFIG_REVISIONS r
       where CONFIG_CONTAINERS.CONFIG_CONTAINER_ID=r.CONFIG_CONTAINER_ID);

--
-- Link each host to the hosts config container
--
update HOSTS
      set CONFIG_CONTAINER_ID=
      (select CONFIG_CONTAINER_ID
      from CONFIG_CONTAINERS cc
      where 'ALL_HOSTS'=cc.CONFIG_TYPE);

--
-- All data insertion is done. Set up, update indices and constraints.
--

create index IDX_CONFIG_CONFIG_CONTAINER on CONFIGS (CONFIG_CONTAINER_ID);

alter table CONFIGS
      add constraint FK_CONFIG_CONFIG_CONTAINER
      foreign key (CONFIG_CONTAINER_ID)
      references CONFIG_CONTAINERS (CONFIG_CONTAINER_ID);

create index IDX_CONFIG_CONTAINER_CONFIG_REVISION on CONFIG_CONTAINERS (CONFIG_REVISION_ID);

alter table CONFIG_CONTAINERS
      add constraint FK_CONFIG_CONTAINER_CONFIG_REVISION
      foreign key (CONFIG_REVISION_ID)
      references CONFIG_REVISIONS (REVISION_ID);

create index IDX_CONFIG_REVISION_CONFIG_CONTAINER on CONFIG_REVISIONS (CONFIG_CONTAINER_ID);

alter table CONFIG_REVISIONS
      add constraint FK_CONFIG_REVISION_CONFIG_CONTAINER
      foreign key (CONFIG_CONTAINER_ID)
      references CONFIG_CONTAINERS (CONFIG_CONTAINER_ID);

create index IDX_HOST_CONFIG_CONTAINER on HOSTS (CONFIG_CONTAINER_ID);

alter table HOSTS
      add constraint FK_HOST_CONFIG_CONTAINER
      foreign key (CONFIG_CONTAINER_ID)
      references CONFIG_CONTAINERS (CONFIG_CONTAINER_ID);
